/*
 Compile: See Makefile in bayeswave directory
 */

/***************************  REQUIRED LIBRARIES  ***************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <math.h>
#include <sys/stat.h>

#include "Constants.h"
#include "Declarations.h"
#include "numrec.h"
#include "BayesLine.h"

/*************  PROTOTYPE DECLARATIONS FOR INTERNAL FUNCTIONS  **************/

void print_help_message(void);
double innerproduct(int imin, int imax, double *a, double *b, double Sn, double *Snf);
void print_time_domain_waveforms(char filename[], double *h, int N, double *Snf, double eta, double Tobs, double fix, int imin, int imax, double tmin, double tmax);
double get_energy(int imin, int imax, double *a, double deltaF);
double get_rhof(int imin, int imax, double *a, double deltaF, double *arg);
double rhonorm(int imax, double *a, double deltaF);
double get_f0(int imax, double *a, double deltaF);
double get_band(int imax, double *a, double deltaF, double f0);
void get_time_domain_waveforms(double *hoft, double *h, int N, double *Snf, double eta, double Tobs, double fix, int imin, int imax, double tmin, double tmax);
double get_rhot(int N, double *a, double deltaT, double *arg);
void whiten_data(int imin, int imax, double *a, double *Snf);
/* ============================  MAIN PROGRAM  ============================ */

int main(int argc, char *argv[])
{

  /*   Variable declaration   */
  int i, d, ifo, imin;//, imax;

  char filename[100];

  /* LAL data structure that will contain all of the data from the frame files */
  LALInferenceRunState *runState = XLALCalloc(1,sizeof(LALInferenceRunState));

  runState->commandLine = LALInferenceParseCommandLine(argc,argv);

  /******************************************************************************/
  /*                                                                            */
  /*  Usage statment if --help                                                  */
  /*                                                                            */
  /******************************************************************************/

  if(LALInferenceGetProcParamVal(runState->commandLine, "--help") || argc==1)
  {
    print_help_message();
    return 0;
  }

  /******************************************************************************/
  /*                                                                            */
  /*  Read/store  data                                                          */
  /*                                                                            */
  /******************************************************************************/

  /*
   Unlike in BayesWaveBurst, we DO NOT want to read and store the IFO data,
   but we have to use LALInferenceReadData to properly setup the runState
   structure and use the injection software.

   This is a hack to fill the runState detector with zeros so that all we store
   in runState is the injected waveform.
  */

  if( LALInferenceGetProcParamVal(runState->commandLine, "--0noise") )
  {
    fprintf(stdout,"Found --0noise in the command line\n");
  } else {
    fprintf(stdout," \n");
    fprintf(stdout,"Error:  You must include --0noise in the command line\n");
    return 0;
  }

  fprintf(stdout, "\n");
  fprintf(stdout, " ======= LALInferenceReadData() ======\n");
  runState->data = LALInferenceReadData(runState->commandLine);
  fprintf(stdout, "\n");


  /******************************************************************************/
  /*                                                                            */
  /*  read/store data                                                           */
  /*                                                                            */
  /******************************************************************************/
  LALInferenceIFOData *dataPtr = NULL;

  dataPtr = runState->data;

  double *SNRinj = NULL;

  int NI;
  int N = dataPtr->timeData->data->length;
  double Tobs = (double)N*dataPtr->timeData->deltaT;

  ifo=0;
  while(dataPtr!=NULL)
  {
    dataPtr = dataPtr->next;
    ifo++;
  }
  NI=ifo;

  double **freqData = dmatrix(0,NI-1,0,N-1);
  double **psd      = dmatrix(0,NI-1,0,N/2-1);

  struct Data *data = malloc(sizeof(struct Data));

  /*
   Unused, but needed for command line and initialization functions
  */
  struct Prior *prior = malloc(sizeof(struct Prior));
  struct Chain *chain = malloc(sizeof(struct Chain));

  parse_command_line(data, chain, prior, runState->commandLine);

  // burstMDC injection
  if(LALInferenceGetProcParamVal(runState->commandLine, "--MDC-cache"))
  {
    SNRinj = malloc((NI+1)*sizeof(double));
    fprintf(stdout, " ==== LALInferenceInjectFromMDC(): started. ====\n");
    InjectFromMDC(runState->commandLine, runState->data, SNRinj);
    fprintf(stdout, " ==== LALInferenceInjectFromMDC(): finished. ====\n\n");
    free(SNRinj);
  }

  // CBC injection
  if(LALInferenceGetProcParamVal(runState->commandLine, "--inj"))
  {
    fprintf(stdout, " ==== LALInferenceInjectInspiralSignal(): started. ====\n");
    LALInferenceInjectInspiralSignal(runState->data,runState->commandLine);
    fprintf(stdout, " ==== LALInferenceInjectInspiralSignal(): finished. ====\n\n");
  }

  //Output Fourier domain data for BayesLine
  dataPtr = runState->data;

  double fmin   = dataPtr->fLow;
  double fmax   = dataPtr->fHigh;
  double deltaT = dataPtr->timeData->deltaT;

  // JBK - About to guess delta f !!
  double deltaF = fmax / (N/2);
  fprintf(stdout, "Max freq is %e", fmax);

  int tsize = dataPtr->timeData->data->length/4;

  double TwoDeltaToverN = 2.0 * deltaT / ((double) dataPtr->timeData->data->length);

  //Copy Fourier domain data & PSD from LAL data types into simple arrays
  ifo=0;
  imin = (int)(fmin*Tobs);
  while(dataPtr!=NULL)
  {
    for(i=0; i<N/2; i++)
    {
      psd[ifo][i] = dataPtr->oneSidedNoisePowerSpectrum->data->data[i]/TwoDeltaToverN;
      if(i<imin)
      {
        freqData[ifo][2*i]   = 0.0;
        freqData[ifo][2*i+1] = 0.0;
      }
      else
      {
        freqData[ifo][2*i]   = 0.5*creal(dataPtr->freqData->data->data[i])/deltaT;
        freqData[ifo][2*i+1] = 0.5*cimag(dataPtr->freqData->data->data[i])/deltaT;
      }
    }
    dataPtr = dataPtr->next;
    ifo++;
  }

  /******************************************************************************/
  /*                                                                            */
  /*  Setup DATA, CHAIN, PRIOR, MODEL structures                                */
  /*                                                                            */
  /******************************************************************************/

  /*
   DATA
   */
  fprintf(stdout, "Initializing data (line 183) ...\n");
  initialize_data(data,freqData,N,tsize,psd,Tobs,NI,fmin,fmax);

  data->detector  = malloc(NI*sizeof(LALDetector*));
  data->epoch     = runState->data->epoch;

  ifo=0;
  dataPtr = runState->data;
  while(dataPtr!=NULL)
  {
    data->detector[ifo] = dataPtr->detector;
    ifo++;
    dataPtr = dataPtr->next;
  }

  /*
   CHAIN
   */
  chain->NC = chain->NCmin;
  allocate_chain(chain,NI);

  /*
   PRIOR
   */
  int Dmax = data->Dmax;
  data->Dmax = 200;

  if(data->noiseOnlyFlag) Dmax=data->Dmax=1;
  int omax = 10;
  if(data->Dmax < omax) omax = data->Dmax;

  // maximum number of sineGaussians
  prior->gmin = ivector(0,data->NI-1);
  prior->gmax = ivector(0,data->NI-1);
  /* Wavelet parameters */
  prior->range = dmatrix(0,4,0,1);

  initialize_priors(data, prior, omax);

  /*
   MODEL
   */
  struct Model *model = malloc(sizeof(struct Model));
  initialize_model(model, data, prior, &chain->seed);

  //Use Welch averaged PSD from LALInferenceReadData
  for(ifo=0; ifo<data->NI; ifo++)
  {
    for(i=0; i<N/2; i++)
    {
      data->SnS[ifo][i] = psd[ifo][i];
    }
  }

  /******************************************************************************/
  /*                                                                            */
  /*  Loop over BayesWaveBurst output chains for post-processing                */
  /*                                                                            */
  /******************************************************************************/

  if(data->bayesLineFlag)
  {

    struct BayesLineParams **bayesline = malloc(NI*sizeof(struct BayesLineParams *));
    data->bayesline = bayesline;


    FILE *splinechain;
    FILE *linechain;

    char burnrows[1000000];
    int count;
    int imin;
    int imax;
    int Nspline;
    //int Nline;

    double *f =malloc(N/2*sizeof(double));
    double *Sn=malloc(N/2*sizeof(double));
    //double *Sl=malloc(N/2*sizeof(double));

    //dataParams *BLdata      = NULL;
    splineParams *spline    = NULL;
    //lorentzianParams *lines = NULL;

    for(i=0; i<N/2; i++) f[i] = (double)(i)/data->Tobs;

    for(ifo=0; ifo<NI; ifo++)
    {

      //Get BayesLine parameter files
      sprintf(filename,"chains/%sclean_1_lorentzchain_ifo%i.dat.0",data->runName,ifo);
      linechain = fopen(filename,"r");

      sprintf(filename,"chains/%sclean_1_splinechain_ifo%i.dat.0",data->runName,ifo);
      splinechain = fopen(filename,"r");


      /*
       THE BROADBAND PSD MODEL
      */

      //Count lines in output files
      count=0;
      while(!feof(splinechain))
      {
        fgets(burnrows,1000000,splinechain);
        count++;
      }
      count--;
      rewind(splinechain);

      //Now burn off all but the last row
      for(i=0; i<count-1; i++)fgets(burnrows,1000000,splinechain);

      //Get dimension of spline model
      fscanf(splinechain,"%i",&Nspline);


      spline = malloc( sizeof(splineParams) );
      create_splineParams(spline,Nspline);

      //Store spline points
      for(i=0; i<Nspline; i++) fscanf(splinechain,"%lg %lg",&spline->points[i],&spline->data[i]);
      imin = (int)(spline->points[0]*data->Tobs);
      imax = (int)(spline->points[Nspline-1]*data->Tobs);

      //Interpolate control points
      /*
       This is tricky:
        - The GSL cubic spline routine throws errors if interpolation
       is outside of control points.
        - Bayesline only fits to frequencies above fmin
        - Some index gymnastics are needed
       */
      for(i=0; i<N/2; i++) data->SnS[ifo][i] = 1.0;

      CubicSplineGSL(Nspline,spline->points,spline->data,imax-imin,f+imin,Sn+imin);

      //Map interpolated points to PSD used later on
      for(i=imin; i<imax; i++) data->SnS[ifo][i]=exp(Sn[i])/data->Tobs/TwoDeltaToverN;

      destroy_splineParams(spline);

      fclose(splinechain);


      /*
       THE SPECTRAL LINES MODEL
       */
      /*
       //Set up BayesLine data structure
       BLdata = malloc(sizeof(dataParams));
       create_dataParams(BLdata,f,N/2);

      //Now burn off all but the last row
      for(i=0; i<count-1; i++)fgets(burnrows,1000000,linechain);

      //Get dimension of line model
      fscanf(linechain,"%i",&Nline);

      lines = malloc( sizeof(lorentzianParams) );
      create_lorentzianParams(lines, Nline);

      //Store line parameters
      for(i=0; i<Nline; i++) fscanf(linechain,"%lg %lg %lg",&lines->f[i],&lines->A[i],&lines->Q[i]);

      //Compute line model
      for(i=0; i<N/2; i++) Sl[i] = 0.0;
      full_spectrum_spline(Sl, Sn, f, BLdata, lines);



      destroy_lorentzianParams(lines);

      free(BLdata);
       */

      fclose(linechain);
    }

    free(f);
    free(Sn);
    //free(Sl);

  }



  fprintf(stdout, "Starting loop over BWB chains (line 243) ...\n");

  //File containing signal-model waveform parameters
  //Was causing segmentation fault. wavechain filename contains the model, not
  //the run name. Should this file opening command be made more general?
  //sprintf(filename,"chains/%s_1_wavechain.dat.0",data->runName);
  sprintf(filename,"chains/signal_1_wavechain.dat.0");
  FILE *signalParams = fopen(filename,"r");

  struct Wavelet *signal = model->signal;

  /*
   Many of the post-processing diagnostics are only meaningful when there is an injected waveform
   and should be skipped otherwise

   //TODO: Add injectionFlag to data structure
  */
  int injectionFlag = 0;
  if(LALInferenceGetProcParamVal(runState->commandLine, "--MDC-cache") || LALInferenceGetProcParamVal(runState->commandLine, "--inj"))
    injectionFlag = 1;

  //User-friendly arrays for PSD, signal, and injected waveform
  //double **psd = dmatrix(0,NI-1,0,N/2-1); PSD matrix is already allocated
  double **hrec = malloc(NI*sizeof(double*));
  double **hinj = malloc(NI*sizeof(double*));
  for(ifo=0; ifo<NI; ifo++)
  {
    hrec[ifo] = malloc(N*sizeof(double));
    hinj[ifo] = malloc(N*sizeof(double));

  }
  //fill hinj with injected waveform
  for(ifo=0; ifo<NI; ifo++) for(i=0; i<N; i++) hinj[ifo][i] = data->s[ifo][i];

  /******************************************************************************/
  /*                                                                            */
  /*  Declare the list of metrics to compute as part                            */
  /*  of the post-processing step.                                              */
  /*                                                                            */
  /******************************************************************************/
  //TODO:  Should each post-processing quantity have its own chain file?
  double overlap, rr, ii, ri, hrhr, hihi, hrhi;
  sprintf(filename,"network_overlap.dat");

  // Open Output files for moments in colored noise
  FILE **momentsOut = malloc(NI*sizeof(FILE*));
  for(ifo=0; ifo<NI; ifo++){
    sprintf(filename,"moments.dat.%i", ifo);
    momentsOut[ifo] = fopen(filename,"w");
    fprintf(momentsOut[ifo], "# overlap energy_inj energy_rec time_inj time_rec duration_inj duration_rec f_inj f_rec band_inj ban_rec \n");
  }

  //  Open output files for moments in whitened data
  FILE **w_momentsOut = malloc(NI*sizeof(FILE*));
  for(ifo=0; ifo<NI; ifo++){
    sprintf(filename,"w_moments.dat.%i", ifo);
    w_momentsOut[ifo] = fopen(filename,"w");
    fprintf(w_momentsOut[ifo], "# overlap energy_inj energy_rec time_inj time_rec duration_inj duration_rec f_inj f_rec band_inj ban_rec \n");
  }

  //  Open output files for reconstructed time domain waveform (colored noise)
  FILE **t_recOut = malloc(NI*sizeof(FILE*));
  for(ifo=0; ifo<NI; ifo++){
    sprintf(filename,"t_recOut.dat.%i", ifo);
    t_recOut[ifo] = fopen(filename,"w");
    fprintf(t_recOut[ifo], "# Each line is h(t) for one sample in the chain");
  }

  //  Open output files for injected time domain waveform (colored noise)
  FILE **t_injOut = malloc(NI*sizeof(FILE*));
  for(ifo=0; ifo<NI; ifo++){
    sprintf(filename,"t_injOut.dat.%i", ifo);
    t_injOut[ifo] = fopen(filename,"w");
    fprintf(t_injOut[ifo], "# Each line is h(t) for one sample in the chain");
  }

  //  Open output files for reconstructed freq domain waveform (colored noise)
  FILE **f_recOut = malloc(NI*sizeof(FILE*));
  for(ifo=0; ifo<NI; ifo++){
    sprintf(filename,"f_recOut.dat.%i", ifo);
    f_recOut[ifo] = fopen(filename,"w");
    fprintf(f_recOut[ifo], "# Each line is h(f) for one sample in the chain: real img real img ...");
  }


  // Write out time and freq. samples
  FILE *timesamp = fopen("timesamp.txt", "w");
  FILE *freqsamp = fopen("freqsamp.txt", "w");
  double t0 = Tobs-2.0;
  double f;
  double t;
  for(i=0; i<N; i++) {
    t = (i*Tobs)/N;
    fprintf(timesamp, "%e \n", t-t0);
  }
  fclose(timesamp);
  for (i=0;i<data->imax;i++) {
    f = i*deltaF;
    fprintf(freqsamp, "%e \n", f);
  }
  fclose(freqsamp);


  mkdir("injected_waveforms",S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
  int mc=0;
  double tempMin  = chain->tempMin;
  // parameters to control how many frames are output: should follow convention in BayesWaveBurst.c
  // TODO: promote these to global?
  int frame=0;
  int frameCount=200;
  int M  = chain->count;
  double frameInterval = (double)M/(double)frameCount;
  // I don't understand the factor of 8 here in the scaling of the time domain reconstructions....
  // double fix = (sqrt(data->Tobs))*sqrt(0.5*(double)N)/8.0;
  double fix = (sqrt(data->Tobs))*sqrt(0.5*(double)N)*4*3.14159;  //JBK - Factor of 4*pi seems to fix normalization

  //work-space for chain diagnostics (periodic time-domain waveforms)
  double *ty=dvector(0,data->N-1);

  int test=1;
  int sampcount=0;
  while(!feof(signalParams))
  {
    sampcount++;
    /******************************************************************************/
    /*                                                                            */
    /*  Get current state of MCMC output into format for BayesWave code library   */
    /*                                                                            */
    /******************************************************************************/

    /*
     SIGNAL MODEL
    */

    //Zero out signal->templates array (holds the linear combination of wavelets)
    for(i=0; i<N; i++) signal->templates[i] = 0.0;

    //Get the number of wavelet basis functions for the current sample
    fscanf(signalParams, "%i", &signal->size);

    //First in the wavechain file are the extrinsic parameters
    for(i=0; i<NE; i++)
      fscanf(signalParams,"%lg", &model->extParams[i]);

    //Compute network projection coefficients
    computeProjectionCoeffs(data, model->projection, model->extParams);

    //Get each wavelets parameters and fill up signal structure
    for(d=1; d<=signal->size; d++)
    {
      signal->index[d]=d;
      for(i=0; i<NG; i++)
        fscanf(signalParams,"%lg ", &signal->intParams[d][i]);

      //Append new wavelet to signal->templates array
      SineGaussian(signal->templates, signal->intParams[d], data->N, 1, data->Tobs);

    }

    //Project geocenter waveform onto network
    waveformProject(data, model->projection, model->extParams, model->response, signal->templates, data->fmin, data->fmax);

    //fill hrec with reconstructed waveform and hinj with injection
    for(ifo=0; ifo<NI; ifo++) for(i=0; i<N; i++) hrec[ifo][i] = model->response[ifo][i];
    for(ifo=0; ifo<NI; ifo++) for(i=0; i<N; i++) hinj[ifo][i] = data->s[ifo][i];

    /*
     PSD MODEL
    */
    //TODO:  Add bayesline support

    //Fill psd with current model PSD
    for(ifo=0; ifo<NI; ifo++) for(i=0; i<N/2; i++) psd[ifo][i] = data->SnS[ifo][i]; //data->SnS holds the "smooth" part of the PSD

    /******************************************************************************/
    /*                                                                            */
    /*  Compute various moments/overlaps to quantify waveform reconstruction      */
    /*                                                                            */
    /******************************************************************************/

    /*
     Overlap between injected and recovered waveforms (for the individal detectors and the network)
     Output whitened injected waveforms to file as was done for the reconstructed wavewforms
    */

    if(injectionFlag)
    {
      // (hrec|hinj)/sqrt( (hrec|hrec) (hinj|hinj) )
      overlap = 0.0; // (hrec|hinj)/sqrt( (hrec|hrec) (hinj|hinj) )
      ii      = 0.0; // (hinj|hinj)
      rr      = 0.0; // (hrec|hrec)
      ri      = 0.0; // (hrec|hinj)

      double white;
      for(ifo=0; ifo<NI; ifo++) {
	for(white=0; white<2; white++) {
	  
	  hihi = innerproduct(imin, N/2, hinj[ifo], hinj[ifo], 1.0, psd[ifo]);
	  hrhr = innerproduct(imin, N/2, hrec[ifo], hrec[ifo], 1.0, psd[ifo]);
	  hrhi = innerproduct(imin, N/2, hrec[ifo], hinj[ifo], 1.0, psd[ifo]);
	  
	  double overlap;
	  overlap = hrhi/sqrt(hihi*hrhr);
	  
	  // Calculate signal energy frequency distribution (aka rhof)
	  // Calculate frequency moments for the reconstructed signal
	  double f0_rec, band_rec, energy_rec;
	  double rhof_rec[data->imax];
	  energy_rec = get_rhof(data->imin, data->imax, hrec[ifo], deltaF, rhof_rec);
	  f0_rec = get_f0(data->imax, rhof_rec, deltaF);
	  band_rec = get_band(data->imax, rhof_rec, deltaF, f0_rec);
	  
	  // Calculate frequency moments for the injected signal
	  double f0_inj, band_inj, energy_inj;
	  double rhof_inj[data->imax];
	  energy_inj = get_rhof(data->imin, data->imax, hinj[ifo], deltaF, rhof_inj);
	  f0_inj = get_f0(data->imax, rhof_inj, deltaF);
	  band_inj = get_band(data->imax, rhof_inj, deltaF, f0_inj);
	  
	  // Calculate the time domain waveform
	  double hoft_rec[N];
	  double hoft_inj[N];
	  get_time_domain_waveforms(hoft_rec, hrec[ifo], N, data->Snf[ifo], 1.0, data->Tobs, fix, data->imin, data->imax, prior->range[0][0], prior->range[0][1]);
	  get_time_domain_waveforms(hoft_inj, hinj[ifo], N, data->Snf[ifo], 1.0, data->Tobs, fix, data->imin, data->imax, prior->range[0][0], prior->range[0][1]);
	  
	  // Calculate the time moments for the reconstructed signal
	  double t0_rec, dur_rec, t_energy_rec;
	  double rhot_rec[N];
	  t_energy_rec = get_rhot(N, hoft_rec, deltaT, rhot_rec);
	  t0_rec =  get_f0(N, rhot_rec, deltaT);
	  dur_rec = get_band(N, rhot_rec, deltaT, t0_rec);
	  
	  // Calculate the time moments for the injected signal
	  double t0_inj, dur_inj, t_energy_inj;
	  double rhot_inj[N];
	  t_energy_inj = get_rhot(N, hoft_inj, deltaT, rhot_inj);
	  t0_inj =  get_f0(N, rhot_inj, deltaT);
	  dur_inj = get_band(N, rhot_inj, deltaT, t0_inj);
	  
	  // fprintf(overlapsOut[ifo], "# energy_inj energy_rec time_inj time_rec duration_inj duration_rec f_inj f_rec band_inj ban_rec \n");
	  if (white == 0) {
	    fprintf(momentsOut[ifo],  "%e  %e   %e   %e  %e   %e   %e   %e  %e  %e  %e \n", overlap, t_energy_inj, t_energy_rec, t0_inj, t0_rec, dur_inj, dur_rec, f0_inj, f0_rec, band_inj, band_rec);
	  } else {
	    fprintf(w_momentsOut[ifo],  "%e %e   %e   %e  %e   %e   %e   %e  %e  %e  %e \n", overlap, t_energy_inj, t_energy_rec, t0_inj, t0_rec, dur_inj, dur_rec, f0_inj, f0_rec, band_inj, band_rec);
	  }
	  // ------ Sanity Checks! -------------
	  if (test == 1) {
	    double freq, t;
	    FILE *rhoout = fopen("rho.txt","w");
	    FILE *t_rhoout = fopen("t_rho.txt","w");
	    FILE *hoftout = fopen("sanity_time.txt","w");
	    double testrho;
	    double energy;
	    // Plot the freq distribution of the singal energy
	    for (i=0;i<data->imax;i++) {
	      freq = deltaF*i;
	      fprintf(rhoout, "%e %e \n", freq, rhof_rec[i]);
	    }
	    fclose(rhoout);
	    
	    // Plot the time distribution of the singal energy
	    for (i=0;i<N;i++) {
	      t = i*deltaT;
	      fprintf(t_rhoout, "%e %e \n", t, rhot_rec[i]);
	    }
	    fclose(t_rhoout);
	    
	    // Plot the time domain waveform
	    for (i=0;i<N;i++) {
	      t = i*deltaT;
	      fprintf(hoftout, "%e %e \n", t, hoft_rec[i]);
	    }
	    fclose(hoftout);
	    
	    // Sanity check that rho is normalized to 1
	    testrho = rhonorm(data->imax, rhof_rec, deltaF);
	    fprintf(stdout, "The normalization of rhof is: %e\n", testrho);
	    testrho = rhonorm(N, rhot_rec, deltaT);
	    fprintf(stdout, "The normalization of rhot is: %e \n", testrho);
	    energy = get_energy(data->imin, data->imax, hrec[ifo], deltaF);
	    fprintf(stdout, "The total signal energy is: %e \n", energy);
	    test=0;
	    fprintf(stdout, "The time domain signal energy is: %e / %e \n", t_energy_inj, t_energy_rec);
	    fprintf(stdout, "The injected/recovered signal energy is: %e / %e  \n", energy_inj, energy_rec);
	    fprintf(stdout, "The injected/recovered central frequency is: %e / %e Hz \n", f0_inj, f0_rec);
	    fprintf(stdout, "The injected/recovered bandwidth is: %e / %e Hz \n", band_inj, band_rec);
	    fprintf(stdout, "The injected/recovered central time is: %e / %e seconds \n", t0_inj, t0_rec);
	    fprintf(stdout, "The injected/recovered duration is: %e / %e seconds \n", dur_inj, dur_rec);
	  }
	  // ------- End Sanity checks ---------

	  // Output some sweet waveforms
	  if((sampcount % 1000) == 0) {
	    if( white == 0 ){
	      for(i=0; i<N; i++) {
		fprintf(t_recOut[ifo], "%e ", hoft_rec[i]);
		fprintf(t_injOut[ifo], "%e ", hoft_inj[i]);
		fprintf(f_recOut[ifo], "%e ", hrec[ifo][i]);
	      }
	      fprintf(t_recOut[ifo], "\n");
	      fprintf(f_recOut[ifo], "\n");
	      fprintf(t_injOut[ifo], "\n");
	    }
	  }
	  
	  // Francesco's code for outputting injected waveform in same way as gnu plot flag
	  // Possible replsced by new big output files
	  //	  if(mc>=frame*frameInterval && mc<M)
	  //{
	  //  sprintf(filename,"injected_waveforms/signal_%i_wave_%d.dat.%i", (int)tempMin, frame, ifo);
	  //  for(i=0; i< N; i++) ty[i] = hinj[ifo][i];
	      // print_time_domain_waveforms(filename, ty, N, data->Snf[ifo], 1.0, data->Tobs, fix, data->imin, data->imax, prior->range[0][0], prior->range[0][1]);
	  //  if(ifo==NI-1) frame++;
	  //}
	  
	  // Whiten the data before the next iteration of the whiten loop
	  whiten_data(imin, N/2, hrec[ifo], psd[ifo]);
	  whiten_data(imin, N/2, hinj[ifo], psd[ifo]);
	}
      }
      
      mc+=chain->cycle;
      
    }
  }



  /******************************************************************************/
  /*                                                                            */
  /*  Free memory, exit cleanly                                                 */
  /*                                                                            */
  /******************************************************************************/

  free(model);
  for(ifo=0;ifo<NI;ifo++){
    fclose(momentsOut[ifo]);
    fclose(w_momentsOut[ifo]);
    fclose(t_recOut[ifo]);
  }

  return 0;

}


void print_help_message(void)
{
  fprintf(stdout,"\n ==== BayesWavePost() USAGE: ====\n\n");
  fprintf(stdout,"REQUIRED:\n");
  fprintf(stdout,"  --ifo IFO              interferometer (H1,L1,V1)\n");
  fprintf(stdout,"  --IFO-flow             minimum frequency (Hz)\n");
  fprintf(stdout,"  --IFO-cache LALAdLIGO  lie about the PSD\n");
  fprintf(stdout,"  --trigtime             GPS trigger time\n");
  fprintf(stdout,"  --srate                sampling rate (Hz)\n");
  fprintf(stdout,"  --seglen               duration of data (s)\n");
  fprintf(stdout,"  --PSDstart             GPS start time for PSD estimation\n");
  fprintf(stdout,"  --PSDlength            duration of PSD estimation length\n");
  fprintf(stdout,"  --dataseed 1234        Lie about random seed for data\n");
  fprintf(stdout,"  --0noise               no noise realization\n");
  fprintf(stdout,"\n");
  fprintf(stdout,"OPTIONAL:\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --- Run parameters   -------------------------------------------------------------\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --runName              run name for output files\n");
  fprintf(stdout,"\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --- Model parameters   -----------------------------------------------------------\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --bayesLine            use BayesLine for PSD model\n");
  fprintf(stdout,"\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --- Priors & Proposals -----------------------------------------------------------\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --Dmax                 maximum number of wavelets per channel (40)\n");
  fprintf(stdout,"\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --- LALInference injection  options  ---------------------------------------------\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --inj injfile.xml      Injection XML file to use\n");
  fprintf(stdout,"  --event N              Event number from Injection XML file to use\n");
  fprintf(stdout,"\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --- Burst MDC injection  options  ------------------------------------------------\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --MDC-channel          IFO1-chan, IFO2-chan, etc\n");
  fprintf(stdout,"  --MDC-cache            IFO1-mdcframe, IFO2-mdcframe, etc\n");
  fprintf(stdout,"  --MDC-prefactor        Rescale injection amplitude (1.0)\n");
  fprintf(stdout,"\n");
  fprintf(stdout,"EXAMPLE:\n");
  fprintf(stdout,"./BayesWavePost --ifo H1 --H1-flow 40 --H1-cache LALAdLIGO --H1-channel LALAdLIGO --trigtime 961627867.167 --srate 1024 --seglen 4 --PSDstart 961627867.167 --PSDlength 1024 --dataseed 1234 --0noise\n");
  fprintf(stdout,"\n");
}

double get_energy(int imin, int imax, double *a, double deltaF)
{
  int i;
  double result;

  // Calculate total signal energy
  result = 0.0;
  for(i=imin; i<imax; i++){
    result += (a[i*2]*a[i*2] + a[i*2+1]*a[i*2+1]);
  }
  result *= 2*deltaF;
  return(result);
}

double get_rhof(int imin, int imax, double *a, double deltaF, double *arg)
{
  // Calculate rho_f (Normalized signal energy frequency distribution.  See moments.pdf eqn 1.5)

  int i;
  double norm;

  // We'll normalize by the total signal energy
  norm = get_energy(imin, imax, a, deltaF);
  for(i=0; i<imax; i++) {
    if (i<imin) {
      arg[i] = 0;
    } else {
      arg[i] = 2*(a[i*2]*a[i*2] + a[i*2+1]*a[i*2+1])/norm;
    }
  }
  // Return value is the total signal energy
  return(norm);
}

double get_rhot(int N, double *a, double deltaT, double *arg)
{
  // Calculate rho_t (Normalized signal energy frequency distribution.  See moments.pdf eqn 1.5)
  int i;
  double norm;

  // We'll normalize by the total signal energy
  norm = 0;
  for(i=0;i<N; i++) {
    norm += a[i]*a[i];
  }
  norm *= deltaT;

  for(i=0; i<N; i++) {
    arg[i] = a[i]*a[i]/norm;
  }
  // Return value is the total signal energy
  return(norm);
}


double rhonorm(int imax, double *a, double deltaF)
{
  // Calculate normalization.  This is a sanity check, since rho is supposed to have a normalization of 1.
  int i;
  double norm;

  norm = 0.0;
  for(i=0; i<imax; i++) {
    norm += (a[i]);
  }
  norm *= deltaF;
  return(norm);
}

double get_f0(int imax, double *a, double deltaF)
{
  // Calculate the central frequency (aka first moment).
  int i;
  double result;

  result = 0.0;
  for(i=0; i<imax; i++) {
    result += (a[i]*i);
  }
  result *= deltaF*deltaF;
  return(result);
}

double get_band(int imax, double *a, double deltaF, double f0)
{
  // Calculate the bandwidth (aka 2nd moment).
  int i;
  double result;

  result = 0.0;
  for(i=0; i<imax; i++) {
    result += ( a[i]*(i*deltaF - f0)*(i*deltaF - f0) );
  }
  result *= deltaF;
  result = sqrt(result);
  return(result);
}

double innerproduct(int imin, int imax, double *a, double *b, double Sn, double *Snf)
{
   int i;
  double arg;

  arg = 0.0;
  for(i=imin; i<imax; i++)
  {
    //arg += (a[i*2]*b[i*2] + a[i*2+1]*b[i*2+1])/Snf[i];
    arg += (a[i*2]*b[i*2] + a[i*2+1]*b[i*2+1]);
  }
  // arg *= 4.0/Sn;
  arg *= 4.0;

  return(arg);
}


void whiten_data(int imin, int imax, double *a, double *Snf)
{
  int i;
  for(i=imin; i<imax; i++) {
    a[i*2] = a[i*2]/sqrt(Snf[i]);
    a[i*2+1] = a[i*2+1]/sqrt(Snf[i]);
  }
}


void print_time_domain_waveforms(char filename[], double *h, int N, double *Snf, double eta, double Tobs, double fix, int imin, int imax, double tmin, double tmax)
{
  /*
   TODO: invFFT comes out time-reversed.
   LAL uses exp(-i2pift) for their Fourier transforms
   Why doesn't drealft(ty-1,N,1) work for inverse FFT?
   In the meantime, hacked invFFT by changing sign of imaginary part
   */
  int i;
  double x,t;
  FILE *waveout = fopen(filename,"w");

  h[0] = 0.0;
  h[1] = 0.0;
  for(i=1; i< N/2; i++)
  {
    if(i>imin && i<imax)
    {
      x = sqrt(Snf[i]*eta);
      h[2*i] /= x;
      h[2*i+1] /= -x;
    }
    else
    {
      h[2*i]=h[2*i+1]=0.0;
    }
  }

  drealft(h-1,N,-1);

  double t0 = Tobs-2.0;

  for(i=0; i<N; i++)
  {
    t = (double)(i)/(double)(N)*Tobs;
    if(t>=tmin && t<tmax) fprintf(waveout,"%e %e\n", t-t0, h[i]/fix);
  }

  fclose(waveout);
}

void get_time_domain_waveforms(double *hoft, double *h, int N, double *Snf, double eta, double Tobs, double fix, int imin, int imax, double tmin, double tmax)
{
  /*
   TODO: invFFT comes out time-reversed.
   LAL uses exp(-i2pift) for their Fourier transforms
   Why doesn't drealft(ty-1,N,1) work for inverse FFT?
   In the meantime, hacked invFFT by changing sign of imaginary part
   */

  // JBK - Why divide by eta?  Shouldn't we multiply by eta?
  int i;
  double x;
  // double x,t;

  hoft[0] = 0.0;
  hoft[1] = 0.0;
  for(i=1; i< N/2; i++)
  {
    if(i>imin && i<imax)
    {
      // x = sqrt(Snf[i]*eta); // Dividing by the Snf should whiten the waveform
      x = sqrt(eta);        //JBK - try making an UNWHITENED time domain waveform
      hoft[2*i] = h[2*i] / x;
      hoft[2*i+1] = h[2*i+1]/(-x);
    }
    else
    {
      hoft[2*i]=hoft[2*i+1]=0.0;
    }
  }

  drealft(hoft-1,N,-1);

  // double t0 = Tobs-2.0;

  for(i=0; i<N; i++) {
    // t = (double)(i)/(double)(N)*Tobs;
    //    if(t>=tmin && t<tmax) fprintf(waveout,"%e %e\n", t-t0, h[i]/fix);
    hoft[i] = hoft[i]/fix;
  }
}
